EXECUTE = "Применить"
REPLY_TO_HLS_AUTHENTICATION = "Ответ на проверку подлинности"
CHANGE_HLS_SECRET = "Изменение пароля"
ADD_OBJECT = "Добавление объекта"
REMOVE_OBJECT = "Удаление объекта"
ACTIVATE_PASSIVE_CALENDAR = "Активировать пассивный календарь"
ADJUST_TO_QUARTER = "Подстройка к ¼ часа"
ADJUST_TO_MEASURING_PERIOD = "Подстройка к измерительному периоду"
ADJUST_TO_MINUTE = "Подстройка к минуте"
ADJUST_TO_PRESET_TIME = "Подстройка к уставке времени"
PRESET_ADJUSTING_TIME = "Задание уставки времени"
SHIFT_TIME = "Сдвиг времени"
REMOTE_DISCONNECT = "Отключение"  # in origin "Удаленное отключение"
REMOTE_RECONNECT = "Подключение"  # in origin "Удаленное включение"
RESET = "Сброс"
IMAGE_TRANSFER_INITIATE = "Инициализация передачи"
IMAGE_BLOCK_TRANSFER = "Передача блока данных"
IMAGE_VERIFY = "Проверка файла"
IMAGE_ACTIVATE = "Активирование файла"
CAPTURE = "Захват"
ENABLE_DISABLE = "Разрешить/запретить (номера)"
INSERT = "Ввести (номер)"
DELETE = "Стереть (номер)"
SECURITY_ACTIVATE = "Усиление безопасности"
GLOBAL_KEY_TRANSFER = "Передача общего ключа"
KEY_TRANSFER = "Передача ключа"
KEY_AGREEMENT = "Ключевые соглашения"
GENERATE_KEY_PAIR = "Генерация пары для ключа"
GENERATE_CERTIFICATE_REQUEST = "Запрос генерации сертификата"
IMPORT_CERTIFICATE = "Импорт сертификата"
EXPORT_CERTIFICATE = "Экспорт сертификата"
REMOVE_CERTIFICATE = "Удаление сертификата"
REQUEST_ACTION = "Запрос действия"
PUSH = "Протолнуть"
# demand register
NEXT_PERIOD = "Следующий период"



from ..data import Data, ic, an, cdt, cst, choices
from ... import enums as enu
from .. import events as ev
from ...types import implementations as impl
from ...version import AppVersion
from ... import ITE_exceptions as exc


class LDN(Data):
    """for ldn"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, impl.octet_string.LDN, classifier=ic.Classifier.STATIC),

    def characteristics_init(self):
        self._cbs_attr_post_init.update(
            {2: lambda: self.collection.set_manufacturer(self.value.contents[:3])}
        )

    @property
    def manufacturer(self) -> bytes:
        return self.value.contents[:3]


class ActiveFirmwareId(Data):
    """for keep version in collection"""
    def characteristics_init(self):
        self._cbs_attr_post_init.update({2: self.__register_value_preset})

    def __register_value_preset(self):
        """need for start control"""
        self.value.register_cb_post_set(self.__handle_value)
        self.__handle_value()

    def __handle_value(self):
        """check new version for more than current collection_version"""
        version = AppVersion.from_str(self.value.contents.decode("utf-8"))
        if (collection_ver := self.collection.server_ver.get(self.logical_name.b)) is not None:
            if version > collection_ver:
                raise exc.NeedUpdate(F"got {version}, but used {collection_ver}. need update server configuration")
            else:
                """ok validation"""
        else:
            """set without check"""


class Unsigned(Data):
    """ with value type: Unsigned """
    A_ELEMENTS = ic.ICAElement(an.VALUE, cdt.Unsigned, classifier=ic.Classifier.DYNAMIC),


class OctetStringDateTime(Data):
    """ with value type: Unsigned """
    A_ELEMENTS = ic.ICAElement(an.VALUE, cst.OctetStringDateTime, classifier=ic.Classifier.DYNAMIC),


class OpeningBody(Data):
    """ RU. 0.0.96.51.0.255. СТО_34.01-5.1-006-2019v3. E 12.1 """
    A_ELEMENTS = ic.ICAElement(an.VALUE, cdt.Unsigned, classifier=ic.Classifier.DYNAMIC),


class SealUnsigned(cdt.Unsigned):
    def get_report(self, with_unit: bool = True) -> str:
        def get_name(value: int):
            """ СПОДЭСv.3 Е.12.5"""
            match value & 0b11:
                case 0: return "Не определено"
                case 1: return "Обжата"
                case 2: return "Взломана"
                case 3: return "Последущие вскрытия"
        return F"Электронные пломбы: корпуса - {get_name(int(self) & 0b11)}, крышки клеммников - {get_name((int(self) >> 2) & 0b11)}"


class SealStatus(Data):
    """ RU. 0.0.96.51.5.255. СТО_34.01-5.1-006-2019v3. E 12.1 """
    A_ELEMENTS = ic.ICAElement(an.VALUE, SealUnsigned, classifier=ic.Classifier.DYNAMIC),


class TerminalsCoverOpeningState(Data):
    """ RU. 0.0.96.51.1.255. СТО_34.01-5.1-006-2019v3. E 12.2 """
    A_ELEMENTS = ic.ICAElement(an.VALUE, cdt.Unsigned, classifier=ic.Classifier.DYNAMIC),


class BitMapData(cdt.Structure):
    start_time: cst.OctetStringDateTime
    stop_time: cst.OctetStringDateTime
    bitmap_data: cdt.OctetString


class ITEBitMap(Data):
    """ITE 0.128.96.13.1.255. Use for send struct lcd screen bitmap(BMP) with start/stop period to server"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, BitMapData, classifier=ic.Classifier.STATIC),


class ChannelNumberValue(cdt.Unsigned):
    @property
    def channel(self) -> enu.ChannelNumber:
        return enu.ChannelNumber(int(self) & 0b0000_0111)

    @channel.setter
    def channel(self, value: enu.ChannelNumber):
        self.set((int(self) & 0b1111_1000) | value)

    @property
    def interface(self) -> enu.Interface:
        return enu.Interface((int(self) & 0b1111_1000) >> 3)

    @interface.setter
    def interface(self, value: enu.Interface):
        self.set((int(self) & 0b0001_1111) | (value << 3))

    def get_report(self, with_unit: bool = True) -> str:
        return F"Номер канала связи: {self.channel.name}, Тип интерфейса: {self.interface.name}"

    def __str__(self):
        return self.report


class CommunicationPortParameter(Data):
    """ RU. 0.0.96.12.4.255. СТО_34.01-5.1-006-2019v3. 13.10. Определение номера порта по которому установлено соединение"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, ChannelNumberValue, default=enu.ChannelNumber.OPTO_P1 + (enu.Interface.OPTO << 3), classifier=ic.Classifier.DYNAMIC),

    @property
    def value(self) -> ChannelNumberValue:
        """override returned type"""
        return self.get_attr(2)


class AnyDateTime(Data):
    """for a-anotation DLMS UA 1000-1 Ed. 14 Table 60"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, choices.any_date_time, classifier=ic.Classifier.STATIC),


class VoltageEventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.voltage_events.get_report(int(self))


class SPODES3VoltageEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.2 События, связанные с напряжением"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, VoltageEventValues, classifier=ic.Classifier.STATIC),


class CurrentEventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.current_events.get_report(int(self))


class SPODES3CurrentEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.3 События, связанные с током"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, CurrentEventValues, classifier=ic.Classifier.STATIC),


class CommutationEventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.commutation_events.get_report(int(self))


class SPODES3CommutationEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.4 События, связанные с вкл./выкл. ПУ, коммутации реле нагрузки"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, CommutationEventValues, classifier=ic.Classifier.STATIC),


class ProgrammingEventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.programming_events.get_report(int(self))


class SPODES3ProgrammingEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.5 События программирования параметров ПУ"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, ProgrammingEventValues, classifier=ic.Classifier.STATIC),


class ExternalEventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.external_impact_events.get_report(int(self))


class SPODES3ExternalEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.6 События внешних воздействий"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, ExternalEventValues, classifier=ic.Classifier.STATIC),


class CommunicationEventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.communication_events.get_report(int(self))


class SPODES3CommunicationEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.7 Коммуникационные события"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, CommunicationEventValues, classifier=ic.Classifier.STATIC),


class AccessEventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.access_events.get_report(int(self))


class SPODES3AccessEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.8 События контроля доступа"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, AccessEventValues, classifier=ic.Classifier.STATIC),


class SelfDiagnosticEventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.self_diagnostics_events.get_report(int(self))


class SPODES3SelfDiagnosticEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.9 Коды событий для журнала самодиагностики"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, SelfDiagnosticEventValues, classifier=ic.Classifier.STATIC),


class ReactivePowerEventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.reactive_power_events.get_report(int(self))


class SPODES3ReactivePowerEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.10 События по превышению реактивной мощности"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, ReactivePowerEventValues, classifier=ic.Classifier.STATIC),


class PowerQuality2EventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.power_quality_status_2.get_report(int(self))


class SPODES3PowerQuality2Event(Data):
    """СТО_34.01-5.1-006-2019v3 E.1 Статус качества сети (журнал качества сети)"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, PowerQuality2EventValues, classifier=ic.Classifier.STATIC),


class PowerQuality1EventValues(cdt.LongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.power_quality_status_1.get_report(int(self))


class SPODES3PowerQuality1Event(Data):
    """СТО_34.01-5.1-006-2019v3 E.2 Статус качества сети (профиль суточных показаний)"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, PowerQuality1EventValues, classifier=ic.Classifier.STATIC),


# KPZ implements
class KPZ1VoltageEventValues(cdt.DoubleLongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.voltage_events.get_report(int(self))


class KPZ1SPODES3VoltageEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.2 События, связанные с напряжением with bag in value type"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, KPZ1VoltageEventValues, classifier=ic.Classifier.STATIC),


class KPZ1CurrentEventValues(cdt.DoubleLongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.current_events.get_report(int(self))


class KPZ1SPODES3CurrentEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.3 События, связанные с током"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, KPZ1CurrentEventValues, classifier=ic.Classifier.STATIC),


class KPZ1CommutationEventValues(cdt.DoubleLongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.commutation_events.get_report(int(self))


class KPZ1SPODES3CommutationEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.4 События, связанные с вкл./выкл. ПУ, коммутации реле нагрузки"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, KPZ1CommutationEventValues, classifier=ic.Classifier.STATIC),


class KPZ1ProgrammingEventValues(cdt.DoubleLongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.programming_events.get_report(int(self))


class KPZ1SPODES3ProgrammingEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.5 События программирования параметров ПУ"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, KPZ1ProgrammingEventValues, classifier=ic.Classifier.STATIC),


class KPZ1ExternalEventValues(cdt.DoubleLongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.external_impact_events.get_report(int(self))


class KPZ1SPODES3ExternalEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.6 События внешних воздействий"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, KPZ1ExternalEventValues, classifier=ic.Classifier.STATIC),


class KPZ1CommunicationEventValues(cdt.DoubleLongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.communication_events.get_report(int(self))


class KPZ1SPODES3CommunicationEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.7 Коммуникационные события"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, KPZ1CommunicationEventValues, classifier=ic.Classifier.STATIC),


class KPZ1AccessEventValues(cdt.DoubleLongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.access_events.get_report(int(self))


class KPZ1SPODES3AccessEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.8 События контроля доступа"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, KPZ1AccessEventValues, classifier=ic.Classifier.STATIC),


class KPZ1SelfDiagnosticEventValues(cdt.DoubleLongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.self_diagnostics_events.get_report(int(self))


class KPZ1SPODES3SelfDiagnosticEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.9 Коды событий для журнала самодиагностики"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, KPZ1SelfDiagnosticEventValues, classifier=ic.Classifier.STATIC),


class KPZ1ReactivePowerEventValues(cdt.DoubleLongUnsigned):
    def get_report(self, with_unit: bool = True) -> str:
        return ev.reactive_power_events.get_report(int(self))


class KPZ1SPODES3ReactivePowerEvent(Data):
    """СТО_34.01-5.1-006-2019v3 Д.10 События по превышению реактивной мощности"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, KPZ1ReactivePowerEventValues, classifier=ic.Classifier.STATIC),


class SPODES3SPODESVersionValue(cdt.OctetString):
    def __init__(self, value="332e30"):
        super(SPODES3SPODESVersionValue, self).__init__(value)
        match AppVersion.from_str(self.contents.decode("utf-8")):
            case AppVersion(0, 0, 0): raise ValueError(F"got invalid SPODES VERSION VALUE with: {self}")
            case AppVersion(_, _, None): """valid version"""
            case _: raise ValueError(F"got invalid SPODES VERSION VALUE with: {self}")


class SPODES3SPODESVersion(Data):
    """СТО_34.01-5.1-006-2019v3 Г.1 Примечание 2"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, SPODES3SPODESVersionValue, classifier=ic.Classifier.STATIC),



from .__class_init__ import *
from ..types.implementations import integers
from itertools import chain


class ControlState(cdt.Enum, elements=(0, 1, 2)):
    """ Shows the internal state of the disconnect control object. """


class ControlMode(cdt.Enum, elements=tuple(chain(range(7), range(129, 135)))):
    """ Configures the behaviour of the disconnect control object for all triggers. Local disconnection is always possible.
    To suppress local disconnection, the corresponding trigger must be inhibited. """

    def get_letters(self) -> str:
        """return transition litters"""
        match int(self):
            case 0:   return ""
            case 1:   return "bcfgde"
            case 2:   return "bcfgae"
            case 3:   return "bcgde"
            case 4:   return "bcgae"
            case 5:   return "bcfgdeh"
            case 6:   return "bcgdeh"
            case 129: return "bcfmde"
            case 130: return "bcfmae"
            case 131: return "bcmde"
            case 132: return "bcma"
            case 133: return "bcmak"
            case 134: return "bcmsdep"
            case _:   raise ValueError(F"unknown {self=}")


class OutputState(cdt.Boolean):
    """ Shows the actual physical state of the disconnect unit, i.e. if an electricity breaker or a gas valve is open or closed. TRUE = connected, FALSE = disconnected.
    In electricity metering, the supply is connected when the disconnector device is closed. In gas and water metering, the supply is connected when the valve is open """
    def __str__(self):
        return en.DISCONNECTED if self.contents == b'\x00' else en.CONNECTED


class DisconnectControl(ic.COSEMInterfaceClasses):
    """ Instances of the Disconnect control interface class manage an internal or external disconnect unit of the meter (e.g. electricity breaker,
    gas valve) in order to connect or disconnect, partly or entirely, the premises of the consumer.
    Disconnect and reconnect can be requested:
        • Remotely, via a communication channel: remote_disconnect, remote_reconnect;
        • Manually, using e.g. a push button: manual_disconnect, manual_reconnect;
        • Locally, by a function of the meter, e.g. limiter, prepayment: local_disconnect. Local reconnection is not possible: reconnection after a
          local_disconnect always requires a manual intervention.
    The possible states and state transitions of the Disconnect control interface class are shown in DLMS UA 1000-1 Ed. 12.0 table 25.The Disconnect
    control object doesn't feature a memory, i.e. any commands are executed immediately
                                States:
    State number | State name            | State description
    0            |Disconnected           |The output_state is set to FALSE and the consumer is disconnected.
    1            |Connected              |The output_state is set to TRUE and the consumer is connected.
    2            |Ready_for_reconnection |The output_state is set to FALSE and the consumer is disconnected.Reconnection requires manual intervention.
                            State transitions:
    Transition |Transition name  | State description
    a          |remote_reconnect |Moves the Disconnect control object from the Disconnected (0) state directly to the Connected (1) state without
                                  manual intervention
    b          |remote_disconnect|Moves the Disconnect control object from the Connected (1) state to the Disconnected (0) state
    c          |remote_disconnect| Moves the Disconnect control object from the Ready_for_reconnection (2) state to the Disconnected (0) state
    d          |remote_reconnect | Moves the Disconnect control object from the Disconnected (0) state to the Ready_for_reconnection (2) state
                                   From this state, it is possible to move to the Connected (2) state via the manual_reconnect transition (e)
    e          |manual_reconnect | Moves the Disconnect control object from the Ready_for_connection (2) state to the Connected (1) state
    f          |manual_disconnect| Moves the Disconnect control object from the Connected (1) state to the Ready_for_connection (2) state From this
                                   state, it is possible to move back to the Connected (2) state via the manual_reconnect transition (e)
    g          |local_disconnect | Moves the Disconnect control object from the Connected (1) state to the Ready_for_connection (2) state From this
                                   state, it is possible to move back to the Connected (2) state via the manual_reconnect transition (e)
                                   NOTE Transitions f) and g) are essentially the same, but their trigger is different."""
    NAME = cn.DISCONNECT_CONTROL
    CLASS_ID = ClassID.DISCONNECT_CONTROL
    VERSION = Version.V0
    A_ELEMENTS = (ic.ICAElement(an.OUTPUT_STATE, OutputState),
                  ic.ICAElement(an.CONTROL_STATE, ControlState),
                  ic.ICAElement(an.CONTROL_MODE, ControlMode))
    M_ELEMENTS = (ic.ICMElement(mn.REMOTE_DISCONNECT, integers.Only0),
                  ic.ICMElement(mn.REMOTE_RECONNECT, integers.Only0))

    def characteristics_init(self):
        """nothing do it"""

    @property
    def output_state(self) -> cdt.Boolean:
        return self.get_attr(2)

    @property
    def control_state(self) -> ControlState:
        return self.get_attr(3)

    @property
    def control_mode(self) -> ControlMode:
        return self.get_attr(4)

    @property
    def remote_disconnect(self) -> cdt.Integer:
        return self.get_meth(1)

    @property
    def remote_reconnect(self) -> cdt.Integer:
        return self.get_meth(2)



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection, overview
from src.DLMS_SPODES.cosem_interface_classes import implementations as impl
from src.DLMS_SPODES.version import AppVersion


class TestType(unittest.TestCase):

    def test_get_index_and_attr(self):
        obj = collection.Data("0.0.0.1.0.255")
        obj.set_attr(2, cdt.Unsigned(8).encoding)
        self.assertEqual(2, len(list(obj.get_index_with_attributes())), "return amount")




from ..__class_init__ import *
from ...types import choices
from ...types.implementations import integers


class DemandRegister(ic.COSEMInterfaceClasses):
    """DLMS UA 1000-1 Ed 14 4.3.4.Demand register"""
    NAME = cn.DEMAND_REGISTER
    CLASS_ID = ClassID.DEMAND_REGISTER
    VERSION = Version.V0
    A_ELEMENTS = (ic.ICAElement(an.CURRENT_AVERAGE_VALUE, choices.register, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement(an.LAST_AVERAGE_VALUE, choices.register, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement(an.SCALER_UNIT, cdt.ScalUnitType),
                  ic.ICAElement(an.STATUS, choices.extended_register, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement(an.CAPTURE_TIME, cst.OctetStringDateTime, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement(an.START_TIME_CURRENT, cst.OctetStringDateTime, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement(an.PERIOD, cdt.DoubleLongUnsigned, min=1),
                  ic.ICAElement(an.NUMBER_OF_PERIODS, cdt.LongUnsigned, min=1, default=1))
    M_ELEMENTS = ic.ICMElement(mn.RESET, integers.Only0),
    scaler_unit_not_settable: bool

    @property
    def current_average_value(self) -> choices.RegisterValues:
        return self.get_attr(2)

    @property
    def last_average_value(self) -> choices.RegisterValues:
        return self.get_attr(3)

    @property
    def scaler_unit(self) -> cdt.ScalUnitType:
        return self.get_attr(4)

    @property
    def status(self) -> choices.ExtendedRegisterValues:
        return self.get_attr(5)

    @property
    def capture_time(self) -> cst.OctetStringDateTime:
        return self.get_attr(6)

    @property
    def start_time_current(self) -> cst.OctetStringDateTime:
        return self.get_attr(7)

    @property
    def period(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(8)

    @property
    def number_of_periods(self) -> cdt.LongUnsigned:
        return self.get_attr(9)

    def characteristics_init(self):
        self._cbs_attr_post_init.update({2: lambda: self.__set_value_data_type(2),
                                         3: lambda: self.__set_value_data_type(3),
                                         4: self.__set_value_scaler_unit})

        self.scaler_unit_not_settable = False
        """ usability scaler unit flag. if True then it not used"""

    def __set_value_data_type(self, attr_index: int):
        """ When instead of a “Data” object a “Register” object is used, (with the scaler_unit attribute not used or with scaler = 0, unit = 255) then the data types allowed for
        the value attribute of the “Data” interface class are allowed. """
        attr_value = self.get_attr(attr_index)
        if self.current_average_value is not None and self.last_average_value is not None and self.current_average_value.TAG != self.last_average_value.TAG:
            self.clear_attr(attr_index)
            raise ValueError(F"got {self.get_attr_element(attr_index).NAME} with: {attr_value=}, expected "
                             F"{cdt.get_common_data_type_from(self.current_average_value.TAG if attr_index == 3 else self.last_average_value.TAG).NAME}")
        match attr_value:
            case cdt.Array() | cdt.CompactArray() | cdt.Structure(): self.set_attr(4, cdt.ScalUnitType(b'\x02\x02\x0f\x00\x16\xff'))
            case cdt.Digital() | cdt.Float():                        attr_value.SCALER_UNIT = self.scaler_unit
            case _:                                                  """ nothing do it """
        match self.scaler_unit:
            case cdt.ScalUnitType(): self.__set_value_scaler_unit()
            case _:                  """ not necessary set Scaler Unit """

    def __set_value_scaler_unit(self):
        match self.current_average_value:
            case cdt.Digital() | cdt.Float() if self.current_average_value.SCALER_UNIT is None:  self.current_average_value.SCALER_UNIT = self.scaler_unit
            case cdt.Digital() | cdt.Float() if self.current_average_value.SCALER_UNIT == self.scaler_unit: """ already set """
            case cdt.Digital() | cdt.Float():                   raise ValueError(F'Got new scaler: {self.scaler_unit} not order with old {self.current_average_value.SCALER_UNIT}')
            case _:                                             """set only for digital"""
        match self.last_average_value:
            case cdt.Digital() | cdt.Float() if self.last_average_value.SCALER_UNIT is None:  self.last_average_value.SCALER_UNIT = self.scaler_unit
            case cdt.Digital() | cdt.Float() if self.last_average_value.SCALER_UNIT == self.scaler_unit: """ already set """
            case cdt.Digital() | cdt.Float():                   raise ValueError(F'Got new scaler: {self.scaler_unit} not order with old {self.last_average_value.SCALER_UNIT}')
            case _:                                             """set only for digital"""



from abc import ABC
from typing import Type
from ... import cosem_interface_classes
from ..register import Register
from ..clock import Clock
from ...relation_to_OBIS import get_name
from ... import ITE_exceptions as exc
from ..__class_init__ import *
from ...types.implementations import integers, arrays, structs

BUFFER = 2
CAPTURE_OBJECTS = 3
CAPTURE_PERIOD = 4
SORT_METHOD = 5
SORT_OBJECT = 6
ENTRIES_IN_USE = 7
PROFILE_ENTRIES = 8


class SortMethod(cdt.Enum, elements=(1, 2, 3, 4, 5, 6)):
    """ If the profile is unsorted, it works as a “first in first out” buffer (it is hence sorted by capturing, and not necessarily by the time
    maintained in the clock object). If the buffer is full, the next call to capture () will push out the first (oldest) entry of the buffer to make
    space for the new entry. If the profile is sorted, a call to capture () will store the new entry at the appropriate position in the buffer, moving
    all following entries and probably losing the least interesting entry. If the new entry would enter the buffer after the last entry and if the
    buffer is already full, the new entry will not be retained at all. """


class CaptureObjects(cdt.Array):
    """ Specifies the list of capture objects """
    TYPE = structs.CaptureObjectDefinition


class FromEntry(cdt.DoubleLongUnsigned, min=1):
    """ Access selector value for selective access to the object_list attribute """


class EntryDescriptor(cdt.Structure):
    """ Only buffer elements corresponding to the entry_descriptor shall be returned in the response.
    NOTE: from_entry and to_entry identify the lines, from_selected_value to_selected_value identify the columns of the buffer to be retrieved. """
    DEFAULT = (1, 0, 1, 0)
    from_entry: FromEntry
    to_entry: cdt.DoubleLongUnsigned
    from_selected_value: cdt.LongUnsigned
    to_selected_value: cdt.LongUnsigned


class AccessSelector(ut.Unsigned8):
    """ Unsigned8 1..2. Default is 2 for read all buffer """
    def __init__(self, value: int | str | ut.Unsigned8 = 2):
        super(AccessSelector, self).__init__(value)
        if int(self) not in (1, 2):
            raise ValueError(F'The {self.__class__.__name__} got {int(self)}, expected 1..2')


class ProfileGeneric(ic.COSEMInterfaceClasses):
    """ The “Profile generic” class defines a generalized concept to store dynamic process values of capture objects. Capture objects are appropriate
    attributes or element of (an) attribute(s) of COSEM objects. The capture objects are collected periodically or occasionally.
    A profile has a buffer to store the captured data. To retrieve only a part of the buffer, either a value range or an entry range may be specified,
    asking to retrieve all entries whose values or entry numbers fall within the given range.
    The list of capture objects defines the values to be stored in the buffer (using the method capture). The list is defined statically to ensure
    homogenous buffer entries (all entries have the same size and structure). If the list of capture objects is modified, the buffer is cleared. If
    the buffer is captured by other “Profile generic” objects, their buffer is cleared as well, to guarantee the homogeneity of their buffer entries.
    The buffer may be defined as sorted by one of the registers or by a clock, or the entries are stacked in a “last in first out” order. So for
    example, it is very easy to build a “maximum demand register” with a one entry deep sorted profile capturing and sorted by a demand register.
    It is just as simple to define a profile retaining the three largest values of some period.
    The size of profile data is determined by three parameters:
        a) the number of entries filled. This will be zero after clearing the profile;
        b) the maximum number of entries to retain. If all entries are filled and a capture () request occurs, the least important entry(according to
           the requested sorting method) will get lost.This maximum number of entries may be specified. Upon changing it, the buffer will be adjusted;
        c) the physical limit for the buffer. This limit typically depends on the objects to capture. The object will reject an attempt of setting
        the maximum number of entries that is larger than physically possible. """
    NAME = cn.PROFILE_GENERIC
    CLASS_ID = ClassID.PROFILE_GENERIC
    VERSION = Version.V1
    scaler_profile_key: bytes | None = None
    """ obis of scaler profile for this profile if need """
    buffer_capture_objects: CaptureObjects
    range_descriptor: Type[cdt.Structure] = None
    attr_descriptor_with_selection: Type[ut.CosemAttributeDescriptorWithSelection] = None
    A_ELEMENTS = (ic.ICAElement(an.BUFFER, arrays.SelectionAccess, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement(an.CAPTURE_OBJECTS, CaptureObjects),
                  ic.ICAElement(an.CAPTURE_PERIOD, cdt.DoubleLongUnsigned),
                  ic.ICAElement(an.SORT_METHOD, SortMethod),
                  ic.ICAElement(an.SORT_OBJECT, structs.CaptureObjectDefinition),
                  ic.ICAElement(an.ENTRIES_IN_USE, cdt.DoubleLongUnsigned, 0, default=0, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement(an.PROFILE_ENTRIES, cdt.DoubleLongUnsigned, 1, default=1))
    M_ELEMENTS = (ic.ICMElement(mn.RESET, integers.Only0),
                  ic.ICMElement(mn.CAPTURE, integers.Only0))

    def characteristics_init(self):
        self.set_attr(BUFFER, None)
        self.buffer.register_cb_preset(lambda _: self.__create_buffer_struct_type())  # value not used for creating struct type

        self._cbs_attr_post_init.update({CAPTURE_OBJECTS: self.__create_buffer_struct_type,
                                         SORT_OBJECT: self.__create_selective_access_descriptor})

        self.buffer_capture_objects = self.capture_objects
        """ objects for buffer. Change with access_selection """

    @property
    def buffer(self) -> arrays.SelectionAccess:
        return self.get_attr(2)

    @property
    def capture_objects(self) -> CaptureObjects:
        return self.get_attr(3)

    @property
    def capture_period(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(4)

    @property
    def sort_method(self) -> SortMethod:
        return self.get_attr(5)

    @property
    def sort_object(self) -> structs.CaptureObjectDefinition:
        return self.get_attr(6)

    @property
    def entries_in_use(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(7)

    @property
    def profile_entries(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(8)

    @property
    def reset(self) -> integers.Only0:
        return self.get_meth(1)

    @property
    def capture(self) -> integers.Only0:
        return self.get_meth(2)

    def get_attr_descriptor(self, value: int) -> ut.CosemAttributeDescriptor | ut.CosemAttributeDescriptorWithSelection:
        """ with selection for object_list. TODO: Copypast AssociationLN"""
        descriptor: ut.CosemAttributeDescriptor = super(ProfileGeneric, self).get_attr_descriptor(value)
        if value == BUFFER and bool(self.collection.current_association.xDLMS_context_info.conformance.decode()[21]):
            if self.attr_descriptor_with_selection is None:
                raise ValueError(F"{self}: get attribute descriptor, not set attribute descriptor with selection. Need initiate capture_objects before")
            else:
                return self.attr_descriptor_with_selection((descriptor, self.buffer.selective_access))
        else:
            return descriptor

    def __create_buffer_struct_type(self):
        """ TODO: more refactoring !!! """
        # rename CaptureObjectDefinition's and adding object if it absense in collection
        if self.capture_objects is None:
            raise ValueError(F"{self}: create buffer Struct type, not set <capture_object> attribute. Need initiate <capture_objects> before")
        for el_value in self.capture_objects:
            el_value: structs.CaptureObjectDefinition
            obj = self.collection.add_if_missing(class_id=ut.CosemClassId(el_value.class_id.contents),
                                                 version=None,
                                                 logical_name=el_value.logical_name)
            self.collection.raise_before(obj, self)
            el_value.set_name(self.collection.get_name_and_type(el_value)[0][-1])
        match self.buffer.selective_access:
            case ut.SelectiveAccessDescriptor() as desc:
                match int(desc.access_selector):
                    # case 0:                                                 self.buffer_capture_objects = self.capture_objects
                    case 1 if len(desc.access_parameters.selected_values) == 0: self.buffer_capture_objects = self.capture_objects
                    case 1:                                                     self.buffer_capture_objects = desc.access_parameters.selected_values
                    case 2:
                        from_selected_value = int(desc.access_parameters.from_selected_value)-1
                        to_selected_value = int(desc.access_parameters.to_selected_value)
                        if to_selected_value == 0:
                            to_selected_value = len(self.capture_objects)
                        self.buffer_capture_objects = self.capture_objects[from_selected_value:to_selected_value]
                    case _ as err:                                                raise ValueError(F'access_selection out of range, got {err}, must be (0..2)')
            case None:
                self.clear_attr(CAPTURE_OBJECTS)
                self._cbs_attr_post_init[CAPTURE_OBJECTS] = self.__create_buffer_struct_type
                raise exc.EmptyObj(F"need set <sort_object> before for {self}")
        buffer_elements: list[cdt.StructElement] = list()
        for el_value in self.buffer_capture_objects:
            names, type_ = self.collection.get_name_and_type(el_value)
            buffer_elements.append(cdt.StructElement(NAME=". ".join(names), TYPE=type_))

        class Entry(cdt.Structure):
            """ The number and the order of the elements of the structure holding the entries is the same as in the definition of the capture_objects.
                The buffer is filled by auto captures or by subsequent calls of the method (capture). The sequence of the entries within the array is ordered
                according to the sort method specified. Default: The buffer is empty after reset.
                REMARK 1 Reading the entire buffer delivers only those entries, which are “in use”.
                REMARK 2 The value of a captured object may be replaced by “null-data” if it can be unambiguously recovered from the previous value
                (e.g. for time: if it can be calculated from the previous value and capture_period; or for a value: if it is equal to the previous value). """
            ELEMENTS = tuple(buffer_elements)

        self.buffer.set_type(Entry)

    def __create_selective_access_descriptor(self):
        """ Available after got sort object. TODO: need rewrite. maybe replace to collection level. Wrong used sort_obj, it can be any element from capture_objects"""
        sort_obj: ic.COSEMInterfaceClasses = self.collection.get_object(self.sort_object.logical_name)
        if sort_obj.CLASS_ID.contents == self.sort_object.class_id.contents:
            value_type: Type[cdt.CommonDataType] = sort_obj.get_attr_data_type(int(self.sort_object.attribute_index))
        else:
            exc.NoObject(F"got {self.sort_object.class_id=}, expected {sort_obj.CLASS_ID=} from collection")

        class RangeDescriptor(cdt.Structure):
            # cb_preset = TODO: make check 'selected_values' from self.capture_objects or
            # cb_post_set = TODO: make check 'selected_values' from self.capture_objects
            DEFAULT = b'\x02\x04\x02\x04\x12\x00\x01\x09\x06\x00\x00\x01\x00\x00\xff\x0f\x02\x12\x00\x00\x09\x0c\x07\xe4\x01\x01\xff\xff\xff\xff\xff\x80\x00\xff' \
                      b'\x09\x0c\x07\xe4\x01\x02\xff\xff\xff\xff\xff\x80\x00\xff\x01\x00'
            restricting_object: structs.CaptureObjectDefinition
            from_value: value_type
            to_value: value_type
            selected_values: CaptureObjects

        class Data(ut.Data):
            restricting_object: structs.CaptureObjectDefinition
            from_value: cdt.SimpleDataType
            to_value: cdt.SimpleDataType
            selected_values: CaptureObjects
            from_entry: FromEntry
            to_entry: cdt.DoubleLongUnsigned
            from_selected_value: cdt.LongUnsigned
            to_selected_value: cdt.LongUnsigned
            ELEMENTS = {1: ut.SequenceElement('range_descriptor', RangeDescriptor),
                        2: ut.SequenceElement('entry_descriptor', EntryDescriptor)}

        class SelectiveAccessDescriptor(ut.SelectiveAccessDescriptor):
            access_selector: AccessSelector
            access_parameters: Data
            ELEMENTS = (ut.SequenceElement('access_selector', AccessSelector),
                        ut.SequenceElement('access_parameters', Data))

        class CosemAttributeDescriptorWithSelection(ut.CosemAttributeDescriptorWithSelection):
            access_selection: SelectiveAccessDescriptor
            ELEMENTS = (ut.SequenceElement('cosem_attribute_descriptor', ut.CosemAttributeDescriptor),
                        ut.SequenceElement('access_selection', SelectiveAccessDescriptor))

        self.attr_descriptor_with_selection = CosemAttributeDescriptorWithSelection
        self.buffer.selective_access = SelectiveAccessDescriptor()

    def get_capture_object_names(self) -> list[str]:
        """ return all capture object names from collection """
        match self.capture_objects:
            case CaptureObjects(): return list(map(lambda definition: get_name(self.collection.get_object(definition).logical_name), self.capture_objects))
            case _:                raise ValueError(F'{self}: Empty capture objects')

    def get_buffer_objects(self) -> list[cosem_interface_classes.cosem_interface_class.COSEMInterfaceClasses]:
        """ get objects of current buffer container """
        return [self.collection.get(obj_def.logical_name.contents) for obj_def in self.buffer_capture_objects]

    # TODO: remove use names created in create_buffer_struct_type
    def get_buffer_elements_names(self) -> list[str]:
        """ get class name + attribute name + unit if possible """
        ret: list[str] = list()
        for capture_obj in self.buffer_capture_objects:
            try:
                capture_obj: structs.CaptureObjectDefinition
                obj: ic.COSEMInterfaceClasses = self.collection.get_object(capture_obj)  # Attention New API!!! remove after test
                match obj, capture_obj.attribute_index.decode():
                    case Register(), 2: additional = F', {obj.scaler_unit.unit if obj.scaler_unit is not None else "?"}'
                    case Register(), 3: additional = F'. {obj.get_attr_element(3).NAME}'
                    case Clock(), 2:    additional = ''
                    case _, index:      additional = F'. {obj.get_attr_element(index).NAME}'
                ret.append(F'{get_name(obj.logical_name)}{additional}')
            except KeyError:
                ret.append('?')
        return ret

    # TODO: Remove by set it in CaptureObjects.append
    def get_scaler_units_profile(self) -> list[cdt.ScalUnitType | None]:
        """ get container of possibles scalers and units from current or scaler_units profiles """
        result: list[cdt.ScalUnitType | None] = list()
        if isinstance(self.scaler_profile_key, bytes):
            scaler_profile = self.collection.get_object(self.scaler_profile_key)
            if not isinstance(scaler_profile.buffer, cdt.Array):
                raise ValueError(F'Загрузите буфер {scaler_profile}. Должен присутствовать в файле конфигурации')
            if len(scaler_profile.buffer) == 0:
                raise ValueError(F'Buffer of {scaler_profile} is empty')
            else:
                unit_scaler_profile = scaler_profile.buffer[0]
                for scaler_unit, capture_object in zip(unit_scaler_profile, self.capture_objects):
                    for buffer_capture_object in self.buffer_capture_objects:
                        if capture_object == buffer_capture_object:
                            match scaler_unit:
                                case cdt.ScalUnitType(): result.append(scaler_unit)
                                case _:                  result.append(None)
                            break
        else:
            for definition in self.buffer_capture_objects:
                definition: structs.CaptureObjectDefinition
                obj = self.collection.get_object(definition)
                match obj, definition.attribute_index.decode():
                    case Register(), 2: result.append(obj.scaler_unit)
                    case _:             result.append(None)
        return result



ASSOCIATION_LN = "Association LN"
ARBITRATOR = "Arbitrator"
PSTN_MODEM_CONFIGURATION = "PSTN modem configuration"
MODEM_CONFIGURATION = "Modem configuration"
SECURITY_SETUP = "Security setup"
ACTIVITY_CALENDAR = "Activity calendar"
CLIENT_SETUP = "Client setup"
CLOCK = "Clock"
DATA = "Data"
DISCONNECT_CONTROL = "Disconnect control"
EXTENDED_REGISTER = "Extended register"
DEMAND_REGISTER = "Demand register"
GPRS_MODEM_SETUP = "GPRS modem setup"
GSM_DIAGNOSTIC = "GSM diagnostic"
IEC_HDLC_SETUP = "IEC HDLC setup"
IEC_LOCAL_PORT_SETUP = "IEC local port setup"
IMAGE_TRANSFER = "Image transfer"
IPV4_SETUP = "IPv4 setup"
LIMITER = "Limiter"
PROFILE_GENERIC = "Profile generic"
PUSH_SETUP = "Push setup"
REGISTER = "Register"
REGISTER_MONITOR = "Register monitor"
SCHEDULE = "Schedule"
SCRIPT_TABLE = "Script table"
SINGLE_ACTION_SCHEDULE = "Single action schedule"
SPECIAL_DAYS_TABLE = "Special days table"
TCP_UDP_SETUP = "TCP UDP setup"
